@@ -26,6 +26,7 @@ import textwrap
 import threading
 import contextlib
 from io import BytesIO, StringIO
+from graphviz.backend import ExecutableNotFound
 import pandas as pd
 import requests
 from requests import Response
@@ -46,6 +47,7 @@ from .util import (
     has_graph_easy,
     interpret_ctype,
     is_jupyter,
+    json_loads,
     merge_ctype,
     safe_opt_num,
     ServerSideError,
@@ -55,9 +57,10 @@ from .types import (
     BlobFilesResponse,
     BlobInit,
     BlobOwner,
+    BlobTypeResponse,
+    BlobURIResponse,
     CacheStats,
     CopyBlob,
-    CSVBlobResponse,
     DagCreate,
     DagDef,
     DagDupResponse,
@@ -73,11 +76,11 @@ from .types import (
     InCursors,
     InstanceStatus,
     JSONBlobAppendResponse,
-    JSONBlobResponse,
     KafkaGroup,
     KafkaMessage,
     KafkaOffsets,
     KafkaThroughput,
+    KafkaTopicNames,
     KafkaTopics,
     KnownBlobs,
     MinimalQueueStatsResponse,
@@ -147,6 +150,7 @@ CUSTOM_NODE_TYPES = {
     "custom_json_to_data",
     "custom_json_join_data",
 }
+NO_RETRY = [METHOD_POST, METHOD_FILE]
 
 
 class AccessDenied(Exception):
@@ -269,6 +273,8 @@ class XYMEClient:
                     raise
                 if not reset_files():
                     raise
+                if method in NO_RETRY:
+                    raise
                 time.sleep(get_retry_sleep())
             retry += 1
 
@@ -297,6 +303,8 @@ class XYMEClient:
                         retry = max_retry
                     raise e
             except RequestException:
+                if method in NO_RETRY:
+                    raise
                 if retry >= max_retry:
                     raise
                 time.sleep(get_retry_sleep())
@@ -351,6 +359,8 @@ class XYMEClient:
                     raise
                 if not reset_files():
                     raise
+                if method in NO_RETRY:
+                    raise
                 time.sleep(get_retry_sleep())
             retry += 1
 
@@ -610,6 +620,26 @@ class XYMEClient:
                 "type": blob_type,
             }, add_namespace=False))["blob"]
 
+    def get_blob_owner(self, blob_uri: str) -> BlobOwner:
+        return cast(BlobOwner, self.request_json(
+            METHOD_GET, "/blob_owner", {
+                "blob": blob_uri,
+            }))
+
+    def set_blob_owner(
+            self,
+            blob_uri: str,
+            dag_id: Optional[str] = None,
+            node_id: Optional[str] = None,
+            external_owner: bool = False) -> BlobOwner:
+        return cast(BlobOwner, self.request_json(
+            METHOD_PUT, "/blob_owner", {
+                "blob": blob_uri,
+                "owner_dag": dag_id,
+                "owner_node": node_id,
+                "external_owner": external_owner,
+            }))
+
     def create_new_dag(
             self,
             username: Optional[str] = None,
@@ -622,6 +652,37 @@ class XYMEClient:
                 "index": index,
             }))["dag"]
 
+    def get_blob_type(self, blob_uri: str) -> BlobTypeResponse:
+        return cast(BlobTypeResponse, self.request_json(
+            METHOD_GET, "/blob_type", {
+                "blob_uri": blob_uri,
+            },
+        ))
+
+    def get_csv_blob(self, blob_uri: str) -> 'CSVBlobHandle':
+        blob_type = self.get_blob_type(blob_uri)
+        if not blob_type["is_csv"]:
+            raise ValueError(f"blob: {blob_uri} is not csv type")
+        return CSVBlobHandle(self, blob_uri, is_full=False)
+
+    def get_model_blob(self, blob_uri: str) -> 'ModelBlobHandle':
+        blob_type = self.get_blob_type(blob_uri)
+        if not blob_type["is_model"]:
+            raise ValueError(f"blob: {blob_uri} is not model type")
+        return ModelBlobHandle(self, blob_uri, is_full=False)
+
+    def get_custom_code_blob(self, blob_uri: str) -> 'CustomCodeBlobHandle':
+        blob_type = self.get_blob_type(blob_uri)
+        if not blob_type["is_custom_code"]:
+            raise ValueError(f"blob: {blob_uri} is not custom code type")
+        return CustomCodeBlobHandle(self, blob_uri, is_full=False)
+
+    def get_json_blob(self, blob_uri: str) -> 'JSONBlobHandle':
+        blob_type = self.get_blob_type(blob_uri)
+        if not blob_type["is_json"]:
+            raise ValueError(f"blob: {blob_uri} is not json type")
+        return JSONBlobHandle(self, blob_uri, is_full=False)
+
     def duplicate_dag(
             self,
             dag_uri: str,
@@ -759,6 +820,12 @@ class XYMEClient:
                 "num_partitions": 1,
             }))
 
+    def get_kafka_error_topic(self) -> str:
+        res = cast(KafkaTopicNames, self.request_json(
+            METHOD_GET, "/kafka_topic_names", {}))["error"]
+        assert res is not None
+        return res
+
     def delete_kafka_error_topic(self) -> KafkaTopics:
         return cast(KafkaTopics, self.request_json(
             METHOD_POST, "/kafka_topics", {
@@ -834,6 +901,46 @@ class XYMEClient:
         return cast(TritonModelsResponse, self.request_json(
             METHOD_GET, "/inference_models", {}))["models"]
 
+    @staticmethod
+    def read_dvc(
+            path: str,
+            repo: str,
+            rev: Optional[str] = "HEAD",
+            warnings_io: Optional[IO[Any]] = sys.stderr) -> Any:
+        """Reading dvc file content from git tracked DVC project.
+
+        Args:
+            path (str):
+                File path to read, relative to the root of the repo.
+            repo (str):
+                specifies the location of the DVC project. It can be a
+                github URL or a file system path.
+            rev (str):
+                Git commit (any revision such as a branch or tag name, or a
+                commit hash). If repo is not a Git repo, this option is
+                ignored. Default: HEAD.
+            warnings_io (optional IO):
+                IO stream where the warning will be printed to
+
+        Returns:
+            the content of the file.
+        """
+        from .util import has_dvc
+
+        if not has_dvc():
+            if warnings_io is not None:
+                warnings_io.write(
+                    "Please install dvc https://dvc.org/doc/install")
+            return None
+
+        import dvc.api
+
+        res = dvc.api.read(path, repo=repo, rev=rev, mode="r")
+        try:
+            return json_loads(res)
+        except json.JSONDecodeError:
+            pass
+        return res
 
 # *** XYMEClient ***
 
@@ -1237,67 +1344,93 @@ class DagHandle:
             self,
             nodes_only: bool,
             allow_unicode: bool,
-            pretty_method: Optional[str] = "accern") -> PrettyResponse:
+            method: Optional[str] = "accern") -> PrettyResponse:
         return cast(PrettyResponse, self._client.request_json(
             METHOD_GET, "/pretty", {
                 "dag": self.get_uri(),
                 "nodes_only": nodes_only,
                 "allow_unicode": allow_unicode,
-                "method": pretty_method,
+                "method": method,
             }))
 
     def pretty(
             self,
             nodes_only: bool = False,
             allow_unicode: bool = True,
-            pretty_method: Optional[str] = "accern",
-            dot_output: Optional[str] = "svg",
-            display: bool = True) -> Optional[str]:
+            method: Optional[str] = "dot",
+            output_format: Optional[str] = "png",
+            display: Optional[IO[Any]] = sys.stdout) -> Optional[str]:
 
         def render(value: str) -> Optional[str]:
-            if display:
-                print(value)
+            if display is not None:
+                display.write(value)
+                display.flush()
                 return None
             return value
 
         graph_str = self._pretty(
             nodes_only=nodes_only,
             allow_unicode=allow_unicode,
-            pretty_method=pretty_method)["pretty"]
-        if pretty_method == "accern":
+            method=method)["pretty"]
+        if method == "accern":
             return render(graph_str)
-        if pretty_method == "dot":
-            from graphviz import Source
+        if method == "dot":
+            try:
+                from graphviz import Source
 
-            graph = Source(graph_str)
-            if dot_output == "dot":
-                return render(graph_str)
-            if dot_output == "svg":
-                svg_str = graph.pipe(format="svg")
-                if display:
-                    if not is_jupyter():
-                        print("Warning: Ipython instance not found.")
-                        print(svg_str)
-                    else:
-                        from IPython.display import display as idisplay
-                        from IPython.display import SVG
-                        idisplay(SVG(svg_str))
-                    return None
-                return svg_str
-            if dot_output == "ascii":
-                if not has_graph_easy():
+                graph = Source(graph_str)
+                if output_format == "dot":
                     return render(graph_str)
+                if output_format == "svg":
+                    svg_str = graph.pipe(format="svg")
+                    if display is not None:
+                        if not is_jupyter():
+                            display.write(
+                                "Warning: Ipython instance not found.\n")
+                            display.write(svg_str)
+                            display.flush()
+                        else:
+                            from IPython.display import display as idisplay
+                            from IPython.display import SVG
+                            idisplay(SVG(svg_str))
+                        return None
+                    return svg_str
+                if output_format == "png":
+                    graph = Source(graph_str)
+                    png_str = graph.pipe(format="png")
+                    if display is not None:
+                        if not is_jupyter():
+                            display.write(
+                                "Warning: Ipython instance not found.\n")
+                            display.write(png_str)
+                            display.flush()
+                        else:
+                            from IPython.display import display as idisplay
+                            from IPython.display import Image
 
-                import subprocess
-                cmd = ["echo", graph_str]
-                p1 = subprocess.Popen(cmd, stdout=subprocess.PIPE)
-                p2 = subprocess.check_output(
-                    ["graph-easy"], stdin=p1.stdout)
-                res = p2.decode("utf-8")
-                return render(res)
-            raise ValueError(
-                "invalid dot output option, use svg, ascii or dot")
-        raise ValueError("invalid dot pretty_method, use accern or dot")
+                            idisplay(Image(png_str))
+                        return None
+                    return png_str
+                if output_format == "ascii":
+                    if not has_graph_easy():
+                        return render(graph_str)
+
+                    import subprocess
+                    cmd = ["echo", graph_str]
+                    p1 = subprocess.Popen(cmd, stdout=subprocess.PIPE)
+                    p2 = subprocess.check_output(
+                        ["graph-easy"], stdin=p1.stdout)
+                    res = p2.decode("utf-8")
+                    return render(res)
+                raise ValueError(
+                    f"invalid format {output_format}, "
+                    "use svg, png, ascii, or dot")
+            except ExecutableNotFound as e:
+                raise RuntimeError(
+                    "use 'brew install graphviz' or use 'method=accern'",
+                ) from e
+        raise ValueError(
+            f"invalid method {method}, use accern or dot")
 
     def pretty_obj(
             self,
@@ -1368,6 +1501,24 @@ class DagHandle:
                 "when": timestamp,
             }))["when"]
 
+    def get_kafka_input_topic(self, postfix: str = "") -> str:
+        res = cast(KafkaTopicNames, self._client.request_json(
+            METHOD_GET, "/kafka_topic_names", {
+                "dag": self.get_uri(),
+                "postfix": postfix,
+                "no_output": True,
+            }))["input"]
+        assert res is not None
+        return res
+
+    def get_kafka_output_topic(self) -> str:
+        res = cast(KafkaTopicNames, self._client.request_json(
+            METHOD_GET, "/kafka_topic_names", {
+                "dag": self.get_uri(),
+            }))["output"]
+        assert res is not None
+        return res
+
     def set_kafka_topic_partitions(
             self,
             num_partitions: int,
@@ -1583,6 +1734,12 @@ class NodeHandle:
         self._state: Optional[int] = None
         self._config_error: Optional[str] = None
 
+    def as_owner(self) -> BlobOwner:
+        return cast(BlobOwner, {
+            "owner_blob": self.get_dag().get_uri(),
+            "owner_node": self.get_id(),
+        })
+
     @staticmethod
     def from_node_info(
             client: XYMEClient,
@@ -1604,7 +1761,7 @@ class NodeHandle:
         return res
 
     def update_info(self, node_info: NodeInfo) -> None:
-        if self._node_id != node_info["id"]:
+        if self.get_id() != node_info["id"]:
             raise ValueError(f"{self._node_id} != {node_info['id']}")
         self._node_name = node_info["name"]
         self._type = node_info["type"]
@@ -1788,32 +1945,43 @@ class NodeHandle:
                 "action": "fix_error",
             }))
 
-    def get_csv_blob(self) -> 'CSVBlobHandle':
-        if self.get_type() != "csv_reader":
-            raise ValueError("node doesn't have csv blob")
+    def get_blob_uri(
+            self, blob_key: str, blob_type: str) -> Tuple[str, BlobOwner]:
         dag = self.get_dag()
-        res = cast(CSVBlobResponse, self._client.request_json(
-            METHOD_GET, "/csv_blob", {
+        res = cast(BlobURIResponse, self._client.request_json(
+            METHOD_GET, "/blob_uri", {
                 "dag": dag.get_uri(),
                 "node": self.get_id(),
+                "key": blob_key,
+                "type": blob_type,
             }))
-        owner: BlobOwner = {
-            "owner_dag": self.get_dag().get_uri(),
-            "owner_node": self.get_id(),
-        }
-        return CSVBlobHandle(self._client, res["csv"], owner)
+        return res["uri"], res["owner"]
 
-    def get_json_blob(self) -> 'JSONBlobHandle':
-        if self.get_type() != "jsons_reader":
-            raise ValueError(
-                f"can not access JSON of {self}, node is not a 'jsons_reader'")
-        dag = self.get_dag()
-        res = cast(JSONBlobResponse, self._client.request_json(
-            METHOD_GET, "/json_blob", {
-                "dag": dag.get_uri(),
-                "node": self.get_id(),
-            }))
-        return JSONBlobHandle(self._client, res["json"], res["count"])
+    def get_csv_blob(self, key: str = "orig") -> 'CSVBlobHandle':
+        uri, owner = self.get_blob_uri(key, "csv")
+        blob = CSVBlobHandle(self._client, uri, is_full=False)
+        blob.set_local_owner(owner)
+        return blob
+
+    def get_json_blob(self, key: str = "jsons_in") -> 'JSONBlobHandle':
+        uri, owner = self.get_blob_uri(key, "json")
+        blob = JSONBlobHandle(self._client, uri, is_full=False)
+        blob.set_local_owner(owner)
+        return blob
+
+    def get_model_blob(
+            self, blob_type: str, key: str = "model") -> 'ModelBlobHandle':
+        uri, owner = self.get_blob_uri(key, blob_type)
+        blob = ModelBlobHandle(self._client, uri, is_full=False)
+        blob.set_local_owner(owner)
+        return blob
+
+    def get_custom_code_blob(
+            self, key: str = "custom_code") -> 'CustomCodeBlobHandle':
+        uri, owner = self.get_blob_uri(key, "custom_code")
+        blob = CustomCodeBlobHandle(self._client, uri, is_full=False)
+        blob.set_local_owner(owner)
+        return blob
 
     def check_custom_code_node(self) -> None:
         if not self.get_type() in CUSTOM_NODE_TYPES:
@@ -1860,37 +2028,6 @@ class NodeHandle:
             },
         ))
 
-    def set_custom_code(self, func: FUNC) -> NodeCustomCode:
-        from RestrictedPython import compile_restricted
-
-        self.check_custom_code_node()
-
-        def fn_as_str(fun: FUNC) -> str:
-            body = textwrap.dedent(inspect.getsource(fun))
-            res = body + textwrap.dedent(f"""
-            result = {fun.__name__}(*data)
-            if result is None:
-                raise ValueError("{fun.__name__} must return a value")
-            """)
-            compile_restricted(res, "inline", "exec")
-            return res
-
-        raw_code = fn_as_str(func)
-        return cast(NodeCustomCode, self._client.request_json(
-            METHOD_PUT, "/custom_code", {
-                "dag": self.get_dag().get_uri(),
-                "node": self.get_id(),
-                "code": raw_code,
-            }))
-
-    def get_custom_code(self) -> NodeCustomCode:
-        self.check_custom_code_node()
-        return cast(NodeCustomCode, self._client.request_json(
-            METHOD_GET, "/custom_code", {
-                "dag": self.get_dag().get_uri(),
-                "node": self.get_id(),
-            }))
-
     def get_user_columns(self, key: str) -> NodeUserColumnsResponse:
         return cast(NodeUserColumnsResponse, self._client.request_json(
             METHOD_GET, "/user_columns", {
@@ -1912,21 +2049,6 @@ class NodeHandle:
             res[key] = df
         return res
 
-    def setup_model(self, obj: Dict[str, Any]) -> ModelSetupResponse:
-        return cast(ModelSetupResponse, self._client.request_json(
-            METHOD_PUT, "/model_setup", {
-                "dag": self.get_dag().get_uri(),
-                "node": self.get_id(),
-                "config": obj,
-            }))
-
-    def get_model_params(self) -> ModelParamsResponse:
-        return cast(ModelParamsResponse, self._client.request_json(
-            METHOD_GET, "/model_params", {
-                "dag": self.get_dag().get_uri(),
-                "node": self.get_id(),
-            }))
-
     def get_def(self) -> NodeDef:
         return cast(NodeDef, self._client.request_json(
             METHOD_GET, "/node_def", {
@@ -1970,6 +2092,8 @@ class BlobHandle:
         self._is_full = is_full
         self._ctype: Optional[str] = None
         self._tmp_uri: Optional[str] = None
+        self._owner: Optional[BlobOwner] = None
+        self._info: Optional[Dict[str, Any]] = None
 
     def is_full(self) -> bool:
         return self._is_full
@@ -1980,9 +2104,27 @@ class BlobHandle:
     def get_uri(self) -> str:
         return self._uri
 
+    def get_path(self, *path: str) -> 'BlobHandle':
+        if self.is_full():
+            raise ValueError(f"URI must not be full: {self}")
+        return BlobHandle(
+            self._client, f"{self._uri}/{'/'.join(path)}", is_full=True)
+
     def get_ctype(self) -> Optional[str]:
         return self._ctype
 
+    def clear_info_cache(self) -> None:
+        self._info = None
+
+    def get_info(self) -> Dict[str, Any]:
+        if self.is_full():
+            raise ValueError(f"URI must not be full: {self}")
+        if self._info is None:
+            info = self.get_path("info.json").get_content()
+            assert info is not None
+            self._info = cast(Dict[str, Any], info)
+        return self._info
+
     def get_content(self) -> Optional[ByteResponse]:
         if not self.is_full():
             raise ValueError(f"URI must be full: {self}")
@@ -2035,19 +2177,29 @@ class BlobHandle:
             }))
 
     def get_owner(self) -> BlobOwner:
-        if self.is_full():
-            raise ValueError(f"URI must not be full: {self}")
-        return cast(BlobOwner, self._client.request_json(
-            METHOD_GET, "/blob_owner", {
-                "blob": self.get_uri(),
-            }))
+        if self._owner is None:
+            self._owner = self._client.get_blob_owner(self.get_uri())
+        return self._owner
+
+    def set_local_owner(self, owner: BlobOwner) -> None:
+        self._owner = owner
+
+    def get_owner_dag(self) -> Optional[str]:
+        owner = self.get_owner()
+        return owner["owner_dag"]
+
+    def get_owner_node(self) -> str:
+        owner = self.get_owner()
+        return owner["owner_node"]
 
     def copy_to(
             self,
             to_uri: str,
-            new_owner: Optional[NodeHandle] = None) -> 'BlobHandle':
+            new_owner: Optional[NodeHandle] = None,
+            external_owner: bool = False) -> 'BlobHandle':
         if self.is_full():
             raise ValueError(f"URI must not be full: {self}")
+
         owner_dag = \
             None if new_owner is None else new_owner.get_dag().get_uri()
         owner_node = None if new_owner is None else new_owner.get_id()
@@ -2056,6 +2208,7 @@ class BlobHandle:
                 "from_uri": self.get_uri(),
                 "owner_dag": owner_dag,
                 "owner_node": owner_node,
+                "external_owner": external_owner,
                 "to_uri": to_uri,
             }))
         return BlobHandle(self._client, res["new_uri"], is_full=False)
@@ -2198,18 +2351,10 @@ class BlobHandle:
 
 
 class CSVBlobHandle(BlobHandle):
-    def __init__(
-            self,
-            client: XYMEClient,
-            uri: str,
-            owner: BlobOwner) -> None:
-        super().__init__(client, uri, is_full=False)
-        self._owner = owner
-
     def finish_csv_upload(self) -> UploadFilesResponse:
         tmp_uri = self._tmp_uri
         assert tmp_uri is not None
-        owner = self._owner
+        owner = self.get_owner()
         args: Dict[str, Optional[Union[str, int]]] = {
             "tmp_uri": tmp_uri,
             "csv_uri": self.get_uri(),
@@ -2268,18 +2413,18 @@ class JSONBlobHandle(BlobHandle):
             self,
             client: XYMEClient,
             uri: str,
-            count: int) -> None:
-        super().__init__(client, uri, is_full=False)
-        self._count = count
+            is_full: bool) -> None:
+        super().__init__(client, uri, is_full)
+        self._count: Optional[int] = None
 
-    def get_count(self) -> int:
+    def get_count(self) -> Optional[int]:
         return self._count
 
     def append_jsons(
             self,
             jsons: List[Any],
             requeue_on_finish: Optional[NodeHandle] = None,
-            ) -> 'JSONBlobHandle':
+            ) -> None:
         obj = {
             "blob": self.get_uri(),
             "jsons": jsons,
@@ -2290,11 +2435,77 @@ class JSONBlobHandle(BlobHandle):
         res = cast(JSONBlobAppendResponse, self._client.request_json(
             METHOD_PUT, "/json_append", obj))
         self._count = res["count"]
-        return self
 
 # *** JSONBlobHandle ***
 
 
+class CustomCodeBlobHandle(BlobHandle):
+    def set_custom_imports(
+            self, modules: List[List[str]]) -> NodeCustomImports:
+        return cast(NodeCustomImports, self._client.request_json(
+            METHOD_PUT, "/custom_imports", {
+                "dag": self.get_owner_dag(),
+                "node": self.get_owner_node(),
+                "modules": modules,
+            }))
+
+    def get_custom_imports(self) -> NodeCustomImports:
+        return cast(NodeCustomImports, self._client.request_json(
+            METHOD_GET, "/custom_imports", {
+                "dag": self.get_owner_dag(),
+                "node": self.get_owner_node(),
+            }))
+
+    def set_custom_code(self, func: FUNC) -> NodeCustomCode:
+        from RestrictedPython import compile_restricted
+
+        def fn_as_str(fun: FUNC) -> str:
+            body = textwrap.dedent(inspect.getsource(fun))
+            res = body + textwrap.dedent(f"""
+            result = {fun.__name__}(*data)
+            if result is None:
+                raise ValueError("{fun.__name__} must return a value")
+            """)
+            compile_restricted(res, "inline", "exec")
+            return res
+
+        raw_code = fn_as_str(func)
+        return cast(NodeCustomCode, self._client.request_json(
+            METHOD_PUT, "/custom_code", {
+                "dag": self.get_owner_dag(),
+                "node": self.get_owner_node(),
+                "code": raw_code,
+            }))
+
+    def get_custom_code(self) -> NodeCustomCode:
+        return cast(NodeCustomCode, self._client.request_json(
+            METHOD_GET, "/custom_code", {
+                "dag": self.get_owner_dag(),
+                "node": self.get_owner_node(),
+            }))
+
+# *** CustomCodeBlobHandle ***
+
+
+class ModelBlobHandle(BlobHandle):
+    def setup_model(self, obj: Dict[str, Any]) -> ModelSetupResponse:
+        return cast(ModelSetupResponse, self._client.request_json(
+            METHOD_PUT, "/model_setup", {
+                "dag": self.get_owner_dag(),
+                "node": self.get_owner_node(),
+                "config": obj,
+            }))
+
+    def get_model_params(self) -> ModelParamsResponse:
+        return cast(ModelParamsResponse, self._client.request_json(
+            METHOD_GET, "/model_params", {
+                "dag": self.get_owner_dag(),
+                "node": self.get_owner_node(),
+            }))
+
+# *** ModelBlobHandle ***
+
+
 class ComputationHandle:
     def __init__(
             self,
@@ -2349,7 +2560,6 @@ class ComputationHandle:
     def __ne__(self, other: object) -> bool:
         return not self.__eq__(other)
 
-
 # *** ComputationHandle ***
 
 
