@@ -229,17 +229,6 @@ TimingResult = TypedDict('TimingResult', {
 InCursors = TypedDict('InCursors', {
     "cursors": Dict[str, int],
 })
-CSVBlobResponse = TypedDict('CSVBlobResponse', {
-    "count": int,
-    "csv": str,
-    "pos": int,
-    "tmp": bool,
-})
-CSVOp = TypedDict('CSVOp', {
-    "count": int,
-    "pos": int,
-    "tmp": bool,
-})
 NamespaceList = TypedDict('NamespaceList', {
     "namespaces": List[str],
 })
@@ -252,10 +241,6 @@ NodeCustomImports = TypedDict('NodeCustomImports', {
 AllowedCustomImports = TypedDict('AllowedCustomImports', {
     "modules": List[str],
 })
-JSONBlobResponse = TypedDict('JSONBlobResponse', {
-    "count": int,
-    "json": str,
-})
 JSONBlobAppendResponse = TypedDict('JSONBlobAppendResponse', {
     "count": int,
 })
@@ -292,6 +277,11 @@ KafkaTopics = TypedDict('KafkaTopics', {
     "topics": Dict[str, Optional[str]],
     "create": bool,
 })
+KafkaTopicNames = TypedDict('KafkaTopicNames', {
+    "input": Optional[str],
+    "output": Optional[str],
+    "error": Optional[str],
+})
 KafkaMessage = TypedDict('KafkaMessage', {
     "messages": Dict[str, str],
 })
@@ -392,3 +382,14 @@ PrettyResponse = TypedDict('PrettyResponse', {
     "pretty": str,
     "nodes": List[DagPrettyNode],
 })
+BlobURIResponse = TypedDict('BlobURIResponse', {
+    "uri": str,
+    "owner": BlobOwner,
+})
+BlobTypeResponse = TypedDict('BlobTypeResponse', {
+    "is_csv": bool,
+    "is_custom_code": bool,
+    "is_json": bool,
+    "is_model": bool,
+    "type": str,
+})
